/*
 * mpeg2dec.c
 * Copyright (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _LARGEFILE_SOURCE	1
#define _LARGEFILE64_SOURCE	1
#define _FILE_OFFSET_BITS	64

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <sched.h>
#include <math.h>

#include "video_out.h"
#include "mpeg2.h"
#include "mfs.h"

#define BUFFER_SIZE 131072
static uint8_t buffer[BUFFER_SIZE];
static FILE * in_file;
static uint32_t frame_counter = 0;

static struct timeval tv_beg, tv_end, tv_start;
static uint32_t elapsed;
static uint32_t total_elapsed;
static uint32_t last_count = 0;
static uint32_t demux_dvd = 0;
static uint32_t demux_pes = 0;
static mpeg2dec_t mpeg2dec;
static vo_output_video_t * video_out;
static char *audio_file;
int mfs_id;
static long skip;
static int iframesonly, ipframesonly;
static double audio_advance;
int highrate;

static void print_fps (int final) 
{
    int fps, tfps, frames;
	
    gettimeofday (&tv_end, NULL);

    if (frame_counter++ == 0) {
	tv_start = tv_beg = tv_end;
    }

    elapsed = (tv_end.tv_sec - tv_beg.tv_sec) * 100 +
	(tv_end.tv_usec - tv_beg.tv_usec) / 10000;
    total_elapsed = (tv_end.tv_sec - tv_start.tv_sec) * 100 +
	(tv_end.tv_usec - tv_start.tv_usec) / 10000;

    if (final) {
	if (total_elapsed) 
	    tfps = frame_counter * 10000 / total_elapsed;
	else
	    tfps = 0;

	fprintf (stderr,"\n%d frames decoded in %d.%02d "
		 "seconds (%d.%02d fps)\n", frame_counter,
		 total_elapsed / 100, total_elapsed % 100,
		 tfps / 100, tfps % 100);

	return;
    }

    if (elapsed < 100)	/* only display every 1.00 seconds */
	return;

    tv_beg = tv_end;
    frames = frame_counter - last_count;

    fps = frames * 10000 / elapsed;			/* 100x */
    tfps = frame_counter * 10000 / total_elapsed;	/* 100x */

#if 0
    fprintf (stderr, "%d frames in %d.%02d sec (%d.%02d fps), "
	     "%d last %d.%02d sec (%d.%02d fps)\033[K\r", frame_counter,
	     total_elapsed / 100, total_elapsed % 100,
	     tfps / 100, tfps % 100, frames, elapsed / 100, elapsed % 100,
	     fps / 100, fps % 100);
#endif

    last_count = frame_counter;
}
 
static void signal_handler (int sig)
{
    print_fps (1);
    signal (sig, SIG_DFL);
    raise (sig);
}

static void print_usage (char * argv[])
{
    int i;

    fprintf (stderr,"usage: %s [-o mode] [-s] file\n"
	     "\t-s\tsystem stream (.vob file)\n"
	     "\t-o\tvideo_output mode\n", argv[0]);

    for (i = 0; video_out_drivers[i]; i++)
	fprintf (stderr, "\t\t\t%s\n", video_out_drivers[i]->name);

    exit (1);
}

static void handle_args (int argc, char * argv[])
{
    int c;
    int i;

    while ((c = getopt (argc,argv,"pds:o:a:m:ijD:H")) != -1) {
	switch (c) {
	case 'o':
	    for (i=0; video_out_drivers[i] != NULL; i++) {
		if (strcmp (video_out_drivers[i]->name, optarg) == 0)
		    video_out = video_out_drivers[i];
	    }
	    if (video_out == NULL) {
		fprintf (stderr, "Invalid video driver: %s\n", optarg);
		print_usage (argv);
	    }
	    break;

	case 'd':
	    demux_dvd = 1;
	    break;

	case 'p':
	    demux_pes = 1;
	    break;

	case 's':
		skip = strtol(optarg, NULL, 0);
		break;

	case 'a':
		audio_file = optarg;
		break;

	case 'm':
		mfs_id = atoi(optarg);
		break;

	case 'i':
		iframesonly = 1;
		break;

	case 'j':
		ipframesonly = 1;
		break;

	case 'D':
		audio_advance = atof(optarg);
		break;

	case 'H':
		highrate = 1;
		break;

	default:
	    print_usage (argv);
	}
    }

    // -o not specified, use a default driver 
    if (video_out == NULL)
	video_out = video_out_drivers[0];

    if (optind < argc) {
	in_file = fopen (argv[optind], "r");
	if (!in_file) {
	    fprintf (stderr, "%s - couldnt open file %s\n", strerror (errno),
		     argv[optind]);
	    exit (1);
	}
    } else
	in_file = stdin;

    mpeg2_init (&mpeg2dec, video_out, NULL);
}

static void ps_loop (void)
{
    static int mpeg1_skip_table[16] = {
	     1, 0xffff,      5,     10, 0xffff, 0xffff, 0xffff, 0xffff,
	0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff
    };

    int num_frames;
    uint8_t * buf;
    uint8_t * end;
    uint8_t * tmp1;
    uint8_t * tmp2;

    buf = buffer;

    do {
	end = buf + fread (buf, 1, buffer + BUFFER_SIZE - buf, in_file);
	buf = buffer;

	while (buf + 4 <= end) {
	    // check start code
	    if (buf[0] || buf[1] || (buf[2] != 0x01)) {
		printf ("missing start code\n");
		exit (1);
	    }

	    switch (buf[3]) {
	    case 0xb9:	/* program end code */
		return;
	    case 0xba:	/* pack header */
		/* skip */
		if ((buf[4] & 0xc0) == 0x40)	/* mpeg2 */
		    tmp1 = buf + 14 + (buf[13] & 7);
		else if ((buf[4] & 0xf0) == 0x20)	/* mpeg1 */
		    tmp1 = buf + 12;
		else if (buf + 5 > end)
		    goto copy;
		else {
		    printf ("weird pack header\n");
		    exit (1);
		}
		if (tmp1 > end)
		    goto copy;
		buf = tmp1;
		break;
	    case 0xe0:	/* video */
		tmp2 = buf + 6 + (buf[4] << 8) + buf[5];
		if (tmp2 > end)
		    goto copy;
		if ((buf[6] & 0xc0) == 0x80)	/* mpeg2 */
		    tmp1 = buf + 9 + buf[8];
		else {	/* mpeg1 */
		    for (tmp1 = buf + 6; *tmp1 == 0xff; tmp1++)
			if (tmp1 == buf + 6 + 16) {
			    printf ("too much stuffing\n");
			    buf = tmp2;
			    break;
			}
		    if ((*tmp1 & 0xc0) == 0x40)
			tmp1 += 2;
		    tmp1 += mpeg1_skip_table [*tmp1 >> 4];
		}
		if (tmp1 < tmp2) {
		    num_frames = mpeg2_decode_data (&mpeg2dec, tmp1, tmp2);
		    while (num_frames--)
			print_fps (0);
		}
		buf = tmp2;
		break;
	    default:
		if (buf[3] < 0xb9) {
		    printf ("looks like a video stream, not system stream\n");
		    exit (1);
		}
		/* skip */
		tmp1 = buf + 6 + (buf[4] << 8) + buf[5];
		if (tmp1 > end)
		    goto copy;
		buf = tmp1;
		break;
	    }
	}

	if (buf < end) {
	copy:
	    /* we only pass here for mpeg1 ps streams */
	    memmove (buffer, buf, end - buf);
	}
	buf = buffer + (end - buf);

    } while (end == buffer + BUFFER_SIZE);
}

static void es_loop (void)
{
    uint8_t * end;
    int num_frames;
		
    do {
	end = buffer + fread (buffer, 1, BUFFER_SIZE, in_file);

	num_frames =
	    mpeg2_decode_data (&mpeg2dec, buffer, end);

	while (num_frames--)
	    print_fps (0);

    } while (end == buffer + BUFFER_SIZE);
}

double timenow(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return tv.tv_sec + tv.tv_usec / 1e6;
}

#define CHUNK_SIZE		(128*1024)

static struct pes_record {
	uint32_t	cmd;
	uint32_t	check;
	uint64_t	time;
} pes_headers[255];

struct bufref {
	uint8_t *buffer;
	uint8_t *start;
	int size;
	double time;
	struct bufref *next;
};

#define GETWORD(p)	(((p)[0] << 24) + ((p)[1] << 16) + ((p)[2] << 8) + (p)[3])

static void pes_loop(void)
{
	int fd = fileno(in_file);
	int nb, nr, i;
	int hdr_ix, hdr_count;
	uint8_t *p;
	FILE *audio;
	unsigned long long mfs_pos = 0, mfs_size = 0;
	int skipping = 1, first_hdr = 0;
	int chunk_ix;
	int nbs = 0, nps = 0;
	int displaying = 1;
	uint acount = 0;
	double atime, prevatime;
	int first_audio = 1;
	uint8_t *free_buffers = 0, *buffer;
	int bufrefs;
	struct bufref *brefs, **brlast, *br;
	double lastftime, t;
	static uint8_t video_flush[] = { 0, 0, 1, 0xff };

	if (mfs_id) {
		mfs_init();
		mfs_size = mfs_fsid_size(mfs_id);
		mfs_size -= mfs_size & (CHUNK_SIZE - 1);
		mfs_readahead(1);
		mfs_pos = (uint64_t) skip * CHUNK_SIZE;
	} else if (skip > 0) {
		lseek(fd, skip * CHUNK_SIZE, SEEK_SET);
	}
	chunk_ix = skip;

	audio = NULL;
	if (audio_file != NULL && !(iframesonly || ipframesonly)) {
		if (audio_file[0] == '|')
			audio = popen(audio_file + 1, "w");
		else
			audio = fopen(audio_file, "w");
	}

	atime = 0;
	prevatime = 0;
	lastftime = 0;
	brefs = 0;
	brlast = &brefs;
	for (;;) {
		if (free_buffers != NULL) {
			buffer = free_buffers;
			free_buffers = *(uint8_t **)buffer;
		} else {
			buffer = malloc(CHUNK_SIZE);
			if (buffer == NULL) {
				fprintf(stderr, "no memory for chunk buffer\n");
				exit(1);
			}
		}
		bufrefs = 0;

		if (mfs_id) {
			if (mfs_pos >= mfs_size)
				break;
			mfs_fsid_pread(mfs_id, buffer, mfs_pos, CHUNK_SIZE);
			mfs_pos += CHUNK_SIZE;
			nb = CHUNK_SIZE;

		} else {
			for (nb = 0; nb < CHUNK_SIZE; nb += nr) {
				nr = read(fd, buffer + nb, CHUNK_SIZE - nb);
				if (nr <= 0)
					break;
			}
			if (nb == 0)
				break;
		}
		p = buffer;
		/* parse the header */
		hdr_ix = 0;
		if (buffer[2] || buffer[3]) {
			hdr_count = 0;
		} else {
			hdr_count = buffer[0];
			first_hdr = buffer[1];
			p += 4;
			for (i = 0; i < hdr_count; ++i) {
				pes_headers[i].cmd = GETWORD(p);
				pes_headers[i].check = GETWORD(p+4);
				pes_headers[i].time = ((uint64_t)(GETWORD(p+8)) << 32) + GETWORD(p+12);
				p += 16;
			}
			nb -= p - buffer;
		}
		for (hdr_ix = 0; hdr_ix < hdr_count; ++hdr_ix) {
			unsigned size, type;
			struct pes_record *pes = &pes_headers[hdr_ix];

			if (pes->check != ~pes->cmd) {
				fprintf(stderr, "%d: BAD PES RECORD: %x %x\n",
					hdr_ix, pes->cmd, pes->check);
				++hdr_ix;
				continue;
			}
			size = pes->cmd >> 12;
			type = pes->cmd & 0xfff;
			if (size > nb) {
				fprintf(stderr, "%d.%d: OUT OF DATA (%u > %d), type %x\n",
					chunk_ix, hdr_ix, size, nb, type);
				break;
			}
			if (skipping && hdr_ix < first_hdr)
				goto cont;
			skipping = 0;
			switch (type) {
			case 0xae0:
			case 0xbe0:
				if (type == 0xae0)
					++nps;
				else
					++nbs;
				displaying = 1;
				if (iframesonly || (ipframesonly && type == 0xbe0)) {
					/* flush out the previous frame */
					i = mpeg2_decode_data(&mpeg2dec, video_flush, video_flush+4);
					while (i--)
						print_fps(0);
					if (iframesonly) {
						skipping = 1;
						nps = nbs = 0;
						if (hdr_ix >= first_hdr)
							first_hdr = 0xff;
					} else
						displaying = 0;
					break;
				}
				t = pes->time / 1e9;
				lastftime += 0.0333667;
				goto playit;
			case 0x2e0:
			case 0x6e0:
			case 0x7e0:
			case 0x8e0:
			case 0xce0:
				/* process video data */
				if (type == 0x8e0) {
					nps = nbs = 0;
					displaying = 1;
					i = pes->time / 1000;
				} else if (!displaying)
					break;
				t = pes->time / 1e9;
			playit:
				t *= 29.97 / 25;
				if (t < lastftime)
					t = lastftime;
				lastftime = t;
				if (t < atime - audio_advance || audio == NULL) {
					i = mpeg2_decode_data(&mpeg2dec, p, p + size);
					while (i--)
						print_fps(0);
				} else {
					/* save up this stuff for later */
					br = (struct bufref *) malloc(sizeof(struct bufref));
					br->buffer = buffer;
					br->start = p;
					br->size = size;
					br->time = t;
					br->next = 0;
					*brlast = br;
					brlast = &br->next;
					++bufrefs;
				}
				break;
			case 0x2c0:
			case 0x3c0:
			case 0x4c0:
				if (audio == NULL)
					break;
				if (acount == 0) {
					int hp;

					if (GETWORD(p) != 0x1c0) {
						if (type == 0x3c0) {
							fprintf(stderr, "%5d.%02d: %x (%d) doesn't start with 1c0\n",
								chunk_ix, hdr_ix, type, size);
							goto dumpit;
						}
						break;
					}

					if (size < 9 || size < 9 + p[8]) {
						fprintf(stderr, "%5d.%02d: %x (%d) is too short\n",
							chunk_ix, hdr_ix, type, size);
						goto dumpit;
					}
					hp = p[8];
					acount = (p[4] << 8) + p[5];
					if (acount < hp + 3) {
						fprintf(stderr, "%5d.%02d: %x (%d) has acount=%d (hp=%d)\n",
							chunk_ix, hdr_ix, type, size, acount, hp);
						goto dumpit;
					}
					if (hp >= 5) {
						uint32_t it = GETWORD(p + 10);
						it = (it & 0xffff) | ((it & ~0x1ffff) >> 1);
						atime = it * 5.5555555e-6;
						if (first_audio)
							prevatime = atime;
						while ((br = brefs) != NULL && br->time < atime - audio_advance) {
							/* play the next bit */
							if (br->time >= prevatime - audio_advance) {
								i = mpeg2_decode_data(&mpeg2dec, br->start,
										      br->start + br->size);
								while (i--)
									print_fps(0);
							}
							if ((brefs = br->next) == NULL)
								brlast = &brefs;
							if (br->buffer == buffer)
								--bufrefs;
							else if (brefs == NULL || br->buffer != brefs->buffer) {
								/* free up that buffer */
								*(uint8_t **)br->buffer = free_buffers;
								free_buffers = br->buffer;
							}
						}
						prevatime = atime;
						first_audio = 0;
					}
					p += 9 + hp;
					nb -= 9 + hp;
					size -= 9 + hp;
					acount -= 3 + hp;
				}
				i = size;
				if (acount < size) {
					fprintf(stderr, "%5d.%02d: %x (%d) audio data ends in record (%d)\n",
						chunk_ix, hdr_ix, type, size, acount);
					i = acount;
				}
				fwrite(p, 1, i, audio);
				fflush(audio);
				acount -= i;
				break;
			default:
			dumpit:
				fprintf(stderr, "%5d.%.2d: %03x %4d %4d.%04d: ",
					chunk_ix, hdr_ix, type, size,
					(int)(pes->time / 1000000000),
					(int)((pes->time / 100000) % 10000));
				for (i = 0; i < size && i < 32; ++i)
					fprintf(stderr, " %.2x", p[i]);
				fprintf(stderr, "\n");
				break;
			}
		cont:
			p += size;
			nb -= size;
		}
		++chunk_ix;
		if (bufrefs == 0) {
			*(uint8_t **)buffer = free_buffers;
			free_buffers = buffer;
		}
	}
	/* flush out the last video */
	while ((br = brefs) != NULL) {
		i = mpeg2_decode_data(&mpeg2dec, br->start, br->start + br->size);
		while (i--)
			print_fps(0);
		if (brefs == NULL || br->buffer != brefs->buffer) {
			/* free up that buffer */
			*(uint8_t **)br->buffer = free_buffers;
			free_buffers = br->buffer;
		}
	}
	i = mpeg2_decode_data(&mpeg2dec, video_flush, video_flush+4);
	while (i--)
		print_fps(0);
	if (audio != NULL)
		fclose(audio);
}

int main (int argc,char *argv[])
{
#if 0
	struct sched_param schedp;
#endif

    printf (PACKAGE"-"VERSION" (C) 2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>\n");

    handle_args (argc,argv);

    signal (SIGINT, signal_handler);

    gettimeofday (&tv_beg, NULL);

#if 0
    schedp.sched_priority = 50;
    sched_setscheduler(getpid(), SCHED_FIFO, &schedp);
#endif

    if (demux_pes)
	pes_loop();
    else if (demux_dvd)
	ps_loop ();
    else
	es_loop ();

    mpeg2_close (&mpeg2dec);
    video_out->close (&video_out);
    print_fps (1);
    return 0;
}
